@@ -44,6 +44,7 @@
 
 #include <event.h>
 
+#include "tree.h"
 #include "http.h"
 #include "html.h"
 #include "db.h"
@@ -52,7 +53,7 @@
 #include "cfg.h"
 
 #define CRAWL		"crawl"
-#define CRAWL_VERSION	"0.2"
+#define CRAWL_VERSION	"0.3"
 
 ssize_t atomicio(ssize_t (*f)(), int, void *, size_t);
 
@@ -69,13 +70,16 @@ regex_t reimgexcl;
 /* Path to directory where the saved images go to */
 char *saveimgdir;
 
+struct timeval tv_timeout;
 int maxdepth = 0;
 
-#define IMG_MINLEN	20000
-#define IMG_MAXLEN	400000
+#define WAIT_DEFAULT	5
 
-int img_minlen = IMG_MINLEN;
-int img_maxlen = IMG_MAXLEN;
+#define MEDIA_MINLEN	20000
+#define MEDIA_MAXLEN	400000
+
+int media_minlen = MEDIA_MINLEN;
+int media_maxlen = MEDIA_MAXLEN;
 
 /* An external filter that does special processing */
 int extmatchw = -1;
@@ -147,6 +151,13 @@ image_saver(struct uri *uri)
 	char *path;
 	char tmp[128], *p = "";
 
+	/* 
+	 * Sometimes we download something that should have been
+	 * html, but is media instead.
+	 */
+	if (uri->save_fd == -1 && !img_permitted(uri->uri))
+		return (-1);
+
 	if (uri->length != -1 && uri->bdlen != uri->length) {
 		snprintf(tmp, sizeof(tmp), " (%4.1f%%/%d)",
 		    (float)uri->bdlen/uri->length*100, uri->length);
@@ -157,13 +168,26 @@ image_saver(struct uri *uri)
 		uri->uri, uri->flags & HTTP_REQUEST_GET ? p : "");
 
 	if (uri->flags & HTTP_REQUEST_HEAD) {
+		int minlen, maxlen;
+
+		if (uri->format == NULL) {
+			minlen = media_minlen;
+			maxlen = media_maxlen;
+		} else {
+			/* Get lengths depending on mime types */
+			minlen = conf_get_num(uri->format, "Min-Length",
+			    media_minlen);
+			maxlen = conf_get_num(uri->format, "Max-Length",
+			    media_maxlen);
+		}
+
 		/* See if it meets our extra constraints */
-		if ((img_minlen != -1 && uri->length < img_minlen) ||
-		    (img_maxlen != -1 && uri->length > img_maxlen))
+		if ((minlen != -1 && uri->length < minlen) ||
+		    (maxlen != -1 && uri->length > maxlen))
 			return (-1);
 
 		/* Re-add request as GET */
-		http_add_dfs(HTTP_REQUEST_GET, uri->uri, uri->depth);
+		http_add(HTTP_REQUEST_GET, uri->uri, uri->depth);
 		return (0);
 	}
 
@@ -179,7 +203,9 @@ image_saver(struct uri *uri)
 			return (-1);
 	}
 
-	atomicio(write, uri->save_fd, uri->body, uri->bdread);
+	if (atomicio(write, uri->save_fd, uri->body, uri->bdread) == -1)
+		return (-1);
+
 	uri->bdread = 0;
 	
 	return (0);
@@ -191,7 +217,7 @@ http_dealimage(char *url, u_short depth)
 	char *path;
 
 	if (!img_permitted(url)) {
-		DFPRINTF(1, (stderr, "Reject: %s\n", url));
+		DFPRINTF(3, (stderr, "Reject: %s\n", url));
 		return;
 	}
 
@@ -204,7 +230,7 @@ http_dealimage(char *url, u_short depth)
 	if (http_mark_seen(url) != 0)
 		return;
 
-	http_add_dfs(HTTP_REQUEST_HEAD, url, depth);
+	http_add(HTTP_REQUEST_HEAD, url, depth);
 }
 
 void
@@ -284,14 +310,14 @@ html_foundanchor(void *arg, char *el, char **attr)
 
 	/* fprintf(stdout, "Anchor: %s\n", normal); */
 	if (!url_permitted(normal)) {
-		DFPRINTF(1, (stderr, "Reject: %s\n", normal));
+		DFPRINTF(3, (stderr, "Reject: %s\n", normal));
 		return;
 	}
 
 	if (http_mark_seen(normal) != 0)
 		return;
 
-	http_add_dfs(HTTP_REQUEST_GET, normal, uri->depth + 1);
+	http_add(HTTP_REQUEST_GET, normal, uri->depth + 1);
 }
 
 int
@@ -317,16 +343,16 @@ void
 http_moved(struct uri *uri, char *location)
 {
 	if (!url_permitted(location)) {
-		DFPRINTF(1, (stderr, "Reject: %s\n", location));
+		DFPRINTF(3, (stderr, "Reject: %s\n", location));
 		return;
 	}
 
 	if (http_mark_seen(location) != 0)
 		return;
 
-	http_add_dfs(uri->flags & HTTP_REQUEST_GET ?
-		     HTTP_REQUEST_GET : HTTP_REQUEST_HEAD,
-		     location, uri->depth);
+	http_add(uri->flags & HTTP_REQUEST_GET ?
+	    HTTP_REQUEST_GET : HTTP_REQUEST_HEAD,
+	    location, uri->depth);
 }
 
 void
@@ -375,7 +401,6 @@ main(int argc, char **argv)
 	extern int optind;
 	char *urlinclude = "http://.*\\.citi\\.umich\\.edu";
 	char *urlexclude = "\\.(ps|gz|c|h|tar|exe|doc|pdf|ppt|txt|diff)$";
-	char *imgmimetype = "image/";
 	char *imginclude = "\\.(jpg|jpeg)";
 	char *imgexclude = "thumbs\\.";
 	char *agent = CRAWL"/"CRAWL_VERSION;
@@ -387,11 +412,12 @@ main(int argc, char **argv)
 
 	progname = argv[0];
 	saveimgdir = ".";
+	timerclear(&tv_timeout);
+	tv_timeout.tv_sec = WAIT_DEFAULT;
 
 	conf_init();
 	CONF_SAVE(urlinclude, conf_get_str("General", "Url-Include"));
 	CONF_SAVE(urlexclude, conf_get_str("General", "Url-Exclude"));
-	CONF_SAVE(imgmimetype, conf_get_str("General", "Img-Mimetype"));
 	CONF_SAVE(imginclude, conf_get_str("General", "Img-Include"));
 	CONF_SAVE(imgexclude, conf_get_str("General", "Img-Exclude"));
 	CONF_SAVE(saveimgdir, conf_get_str("General", "Img-Directory"));
@@ -401,12 +427,24 @@ main(int argc, char **argv)
 	use_robots = conf_get_num("HTTP", "Use-Robots", 1);
 	http_maxconnects = conf_get_num("HTTP", "Connections", HTTP_MAXCONNECTS);
 	maxdepth = conf_get_num("General", "Max-Depth", 0);
+	media_minlen = conf_get_num("General", "Min-Length", MEDIA_MINLEN);
+	media_maxlen = conf_get_num("General", "Max-Length", MEDIA_MAXLEN);
 
-	img_minlen = conf_get_num("Images", "Min-Length", IMG_MINLEN);
-	img_maxlen = conf_get_num("Images", "Max-Length", IMG_MAXLEN);
-
-	while ((ch = getopt(argc, argv, "v:u:e:i:d:m:RA:E:I:c:")) != -1)
+	while ((ch = getopt(argc, argv, "t:v:u:e:i:d:m:RA:E:I:c:")) != -1)
 		switch(ch) {
+		case 't': {
+			char *p;
+			float val;
+			val = strtod(optarg, &p);
+			if (p == NULL && *p != '\0') {
+				usage();
+				exit(1);
+			}
+
+			tv_timeout.tv_sec = val;
+			tv_timeout.tv_usec = (val - (int)val) * 1000000L;
+			break;
+		}
 		case 'v':
 			debug = atoi(optarg);
 			break;
@@ -468,12 +506,13 @@ main(int argc, char **argv)
 	db_setup("crawl.db");
 
 	event_init();
-	http_init(state);
+	/* dns_init has to go before http_init because of signal handlers */
 	dns_init();
+	http_init(state);
 
 	if (resume == NULL) {
 		while (argc) {
-			if (http_add_dfs(HTTP_REQUEST_GET, argv[0], 0) != -1)
+			if (http_add(HTTP_REQUEST_GET, argv[0], 0) != -1)
 				http_mark_seen(argv[0]);
 			
 			argc--;
@@ -483,11 +522,17 @@ main(int argc, char **argv)
 		if (http_restore_state(resume) == -1)
 			exit(1);
 
+	/* Schedule connections */
+	while (http_postevent() != -1)
+		;
+
 	if (strlen(agent))
 		http_setuseragent(agent);
 
 	http_register_dispatch("text/html", html_follower);
-	http_register_dispatch(imgmimetype, image_saver);
+	http_register_dispatch("image/", image_saver);
+	http_register_dispatch("audio/", image_saver);
+	http_register_dispatch("video/", image_saver);
 
 	/* XXX - bad cludge */
 	http_setcallback(http_movecb, http_moved);
@@ -510,6 +555,7 @@ main(int argc, char **argv)
 	/* Print some informative stats */
 	http_print_stats();
 	dns_print_stats();
+	dns_end();
 
 	/* Done with everything */
 	html_freeparser(parser);
