@@ -97,11 +97,14 @@ void
 db_setup(char *file)
 {
 	int flags, mode;
+	BTREEINFO binfo;
 
 	flags = O_RDWR|O_CREAT;
 	mode = S_IRUSR|S_IWUSR;
 
-	db = dbopen(file, flags, mode, DB_HASH, NULL);
+	memset(&binfo, 0, sizeof(binfo));
+	binfo.cachesize = 10000000;
+	db = dbopen(file, flags, mode, DB_BTREE, &binfo);
 	if (db == NULL)
 		errx(1, "dbopen");
 }
