@@ -47,20 +47,29 @@
 
 #include <event.h>
 
+#include "tree.h"
 #include "connection.h"
 #include "http.h"
 #include "db.h"
 #include "robots.h"
+#include "dns.h"
+#include "util.h"
 
 /* Proto-types */
 ssize_t atomicio(ssize_t (*f)(), int, void *, size_t);
+void http_insert_order(struct uri_list *, struct uri_small *);
 void http_connectioncb(int, short, void *);
 int http_isrobotstxt(struct uri *);
+int http_robotstxt(struct uri *);
 
+extern struct timeval tv_timeout;
 extern int debug;
 #define DFPRINTF(x,y)   if (debug >= x) fprintf y
 
-TAILQ_HEAD(uri_list, uri_small) uriqueue;
+/* Two separate queues, mediaqueue gets priority */
+struct uri_list uriqueue;
+struct uri_list mediaqueue;
+
 TAILQ_HEAD(dispatch_list, dispatch) dispatchqueue;
 
 u_char body[32768];
@@ -70,8 +79,12 @@ char *statefile = NULL;
 
 void (*http_movecb)(struct uri *, char *) = NULL;
 
+/* Timeout used for postevent */
+struct event ev_timeout;
+
 int http_maxconnects = HTTP_MAXCONNECTS;
 int connections;
+int dns_connections;
 
 int use_robots = 1;
 /* Set by signal handler if we are supposed to terminate */
@@ -119,7 +132,7 @@ uri_free(struct uri *uri)
 	
 	/* If this is an aborted robots.txt, we treat it as no robots.txt */
 	if (use_robots)
-		http_isrobotstxt(uri);
+		http_robotstxt(uri);
 
 	if (uri->fd != -1) {
 		int n = 0;
@@ -128,11 +141,12 @@ uri_free(struct uri *uri)
 		connections--;
 
 		/* Try to start new connections */
-		if (TAILQ_FIRST(&uriqueue) != NULL) {
-			while (http_postevent() != -1)
-				n++;
-		}
+		while (http_postevent() != -1)
+			n++;
 	}
+
+	if (uri->dns != NULL)
+		dns_unref(uri->dns);
 	
 	if (uri->save_fd != -1)
 		close(uri->save_fd);
@@ -180,12 +194,24 @@ terminate_handler(int sig)
 	errno = s;
 }
 
+void
+http_posteventcb(int fd, short which, void *arg)
+{
+	/* Reschedule events if possible */
+	while (http_postevent() != -1)
+		;
+}
+
 int
 http_init(char *filename)
 {
 	TAILQ_INIT(&uriqueue);
+	TAILQ_INIT(&mediaqueue);
+
 	TAILQ_INIT(&dispatchqueue);
 
+	timeout_set(&ev_timeout, http_posteventcb, NULL);
+
 	if (use_robots)
 		http_robots_init();
 
@@ -319,27 +345,146 @@ http_fetch(struct uri *uri)
 	event_add(&uri->ev, &tv);
 }
 
-int
-http_newconnection(struct uri *uri, char *host, u_short port)
+void
+http_collect(char *name, struct uri_list *dsthead, struct uri_list *srchead)
 {
+	char *host;
+	struct uri_small *tmp, *next;
+
+	for (tmp = TAILQ_FIRST(srchead); tmp; tmp = next) {
+		next = TAILQ_NEXT(tmp, next);
+
+		/* XXX - should not happen */
+		if (http_hostportfile(tmp->uri, &host, NULL, NULL) == -1)
+			continue;
+
+		if (strcasecmp(name, host))
+			continue;
+
+		/* Remove from generic queue, and put on dns queue */
+		TAILQ_REMOVE(srchead, tmp, next);
+		http_insert_order(dsthead, tmp);
+	}
+}
+
+/* A negative dns result, causes us to remove all entries for this host */
+
+void
+http_decollect(char *name, struct uri_list *srchead)
+{
+	char *host;
+	struct uri_small *tmp, *next;
+
+	for (tmp = TAILQ_FIRST(srchead); tmp; tmp = next) {
+		next = TAILQ_NEXT(tmp, next);
+
+		if (http_hostportfile(tmp->uri, &host, NULL, NULL) != -1 &&
+		    strcasecmp(name, host))
+			continue;
+
+		/* Remove from generic queue, and free */
+		TAILQ_REMOVE(srchead, tmp, next);
+		uri_small_free(tmp);
+	}
+}
+
+void
+http_dns_calcdepth(struct dns_entry *dns)
+{
+	struct uri_small *suri;
+	int depth = 0;
+
+	suri = TAILQ_FIRST(&dns->mediaqueue);
+	if (suri != NULL && suri->depth > depth)
+		    depth = suri->depth;
+
+	suri = TAILQ_FIRST(&dns->uriqueue);
+	if (suri != NULL && suri->depth > depth)
+		    depth = suri->depth;
+
+	if (dns->depth != dns->depth)
+		dns_setdepth(dns, depth);
+}
+
+void
+http_newconnection_dnscb(struct addrinfo *ai, struct dns_entry *dns, void *arg)
+{
+	struct uri *uri = arg;
 	struct timeval tv;
+	u_short port;
 	int fd;
 
-	fd = connection_new(host, port);
-	if (fd == -1) {
-		fprintf(stderr, __FUNCTION__
-			": connection_new(%s,%d) failed\n", host, port);
-		return (-1);
+	dns_connections--;
+
+	/* See if we should queue addresses on the dns object */
+	if (dns != NULL) {
+		if (DNS_NEGATIVE(dns)) {
+			DFPRINTF(1, (stderr, __FUNCTION__
+				     ": decollecting %s\n", dns->name));
+			/* the DNS object might have uris queued on it */
+			http_decollect(dns->name, &dns->mediaqueue);
+			http_decollect(dns->name, &dns->uriqueue);
+			/* Search generic list */
+			http_decollect(dns->name, &mediaqueue);
+			http_decollect(dns->name, &uriqueue);
+			goto fail;
+		} else if (timercmp(&dns->creat, &dns->access, =)) {
+			/* First time this dns object has been used */
+			http_collect(dns->name, &dns->mediaqueue, &mediaqueue);
+			http_collect(dns->name, &dns->uriqueue, &uriqueue);
+		}
+
+		http_dns_calcdepth(dns);
 	}
 
-	connections++;
+	/* DNS lookup failed */
+	if (ai == NULL)
+		goto fail;
+
+	/* Complete sillyness of the API */
+	http_hostportfile(uri->uri, NULL, &port, NULL);
+	fd = connection_new(ai, port);
+	if (fd == -1)
+		goto fail;
 
 	event_set(&uri->ev, fd, EV_WRITE, http_connectioncb, uri);
 	timerclear(&tv);
 	tv.tv_sec = HTTP_CONNECT_TIMEOUT;
 	event_add(&uri->ev, &tv);
 
-	return (fd);
+	/* Reference counting for host accesses */
+	if (uri->dns != NULL)
+		dns_unref(uri->dns);
+	dns_ref(dns);
+	uri->dns = dns;
+
+	/* Record next time that we may access this host */
+	gettimeofday(&dns->access, NULL);
+	timeradd(&dns->access, &tv_timeout, &dns->access);
+
+	uri->fd = fd;
+	return;
+
+ fail:
+	connections--;
+	uri_free(uri);
+}
+
+int
+http_newconnection(struct uri *uri, char *host, u_short port)
+{
+	int res;
+
+	connections++;
+	dns_connections++;
+
+	res = dns_resolve_cb(host, port, http_newconnection_dnscb, uri);
+	if (res == -1) {
+		dns_connections--;
+		fprintf(stderr, __FUNCTION__
+		    ": http_newconnection(%s,%d) failed\n", host, port);
+	}
+	return (res);
 }
 
 void
@@ -360,16 +505,13 @@ http_connection_retry(struct uri *uri)
 	connections--;
 
 	http_hostportfile(uri->uri, &host, &port, NULL);
-	uri->fd = http_newconnection(uri, host, port);
-	if (uri->fd == -1)
+	if (http_newconnection(uri, host, port) == -1)
 		goto fail;
 
 	return;
  fail:
 	warnx(__FUNCTION__": failed retry: %s", uri->uri);
 	uri_free(uri);
-
-	http_postevent();
 }
 
 void
@@ -426,7 +568,9 @@ http_hostportfile(char *url, char **phost, u_short *pport, char **pfile)
 
 	url += len;
 
-	strlcpy(host, url, sizeof (host));
+	/* We might overrun */
+	if (strlcpy(host, url, sizeof (host)) >= sizeof(host))
+		return (-1);
 
 	p = strchr(host, '/');
 	if (p != NULL) {
@@ -435,10 +579,12 @@ http_hostportfile(char *url, char **phost, u_short *pport, char **pfile)
 	} else
 		p2 = NULL;
 
-	/* Generate request file */
-	if (p2 == NULL)
-		p2 = "";
-	snprintf(file, sizeof(file), "/%s", p2);
+	if (pfile != NULL) {
+		/* Generate request file */
+		if (p2 == NULL)
+			p2 = "";
+		snprintf(file, sizeof(file), "/%s", p2);
+	}
 
 	p = strchr(host, ':');
 	if (p != NULL) {
@@ -486,13 +632,13 @@ http_restore_state(char *filename)
 
 		p = strsep(&p2, " ");
 		if (p == NULL || p2 == NULL ||
-		    (depth = atoi(p)) == 0)
+		    (depth = atoi(p)) < 0)
 			goto fail;
 
 		if (strncasecmp(p2, HTTP_PREFIX, strlen(HTTP_PREFIX)))
 			goto fail;
 
-		http_add_bfs(type, p2, depth);
+		http_add(type, p2, depth);
 
 		linenum++;
 	}
@@ -508,103 +654,155 @@ http_restore_state(char *filename)
 }
 
 void
-http_savestate(void)
+http_savequeue(FILE *fp, struct uri_list *head, int *num)
 {
 	struct uri_small *uri;
+	int linenum = *num;
+
+	while ((uri = TAILQ_FIRST(head)) != NULL) {
+		TAILQ_REMOVE(head, uri, next);
+
+		fprintf(fp, "%c %d %s\n",
+			uri->flags & HTTP_REQUEST_GET ? 'g' : 'h',
+			uri->depth, uri->uri);
+
+		linenum++;
+	}
+
+	*num = linenum;
+}
+
+void
+http_savestate(void)
+{
+	extern struct dns_list dnsqueue;
+	struct dns_entry *dns;
+
 	FILE *fpstate = NULL;
 	int linenum = 0;
 
+	timeout_del(&ev_timeout);
+
 	fprintf(stdout, "Terminating...\n");
 
 	if ((fpstate = fopen(statefile, "w")) == NULL)
 		err(1, "fopen");
 
-	while ((uri = TAILQ_FIRST(&uriqueue)) != NULL) {
-		TAILQ_REMOVE(&uriqueue, uri, next);
-
-		fprintf(fpstate, "%c %d %s\n",
-			uri->flags & HTTP_REQUEST_GET ? 'g' : 'h',
-			uri->depth, uri->uri);
-
-		linenum++;
+	TAILQ_FOREACH(dns, &dnsqueue, next) {
+		http_savequeue(fpstate, &dns->mediaqueue, &linenum);
+		http_savequeue(fpstate, &dns->uriqueue, &linenum);
 	}
+
+	http_savequeue(fpstate, &mediaqueue, &linenum);
+	http_savequeue(fpstate, &uriqueue, &linenum);
+
 	fclose(fpstate);
 
 	fprintf(stdout, "Terminated with %d saved urls.\n", linenum);
 }
 
+#define HTTP_RBTCHK_PROCEED	0	/* Okay to go ahead */
+#define HTTP_RBTCHK_DENY	1	/* Not okay */
+#define HTTP_RBTCHK_ASKING	2	/* Retrieving robots.txt now */
+#define HTTP_RBTCHK_PENDING	3	/* Still asking */
+
 int
-http_postevent(void)
+http_robot_check(struct uri_list *head, struct uri_small *suri,
+    char *host, short port)
+{
+	struct http_robots *htrobot;
+
+	if ((htrobot = http_robots_find(host, port)) == NULL) {
+		/* Generates a new connection */
+		if (http_robots_get(host, port) != -1)
+			return (HTTP_RBTCHK_ASKING);
+	} else {
+		if (htrobot->flags & HTROBOT_PENDING)
+			return (HTTP_RBTCHK_PENDING);
+		if (!http_robots_allow(htrobot, suri->uri)) {
+			fprintf(stderr, __FUNCTION__
+			    ": robots diallowed url: %s\n",
+			    suri->uri);
+				
+			TAILQ_REMOVE(head, suri, next);
+			STATS_DECFIFO;
+
+			uri_small_free(suri);
+			return (HTTP_RBTCHK_DENY);
+		}
+	}
+
+	return (HTTP_RBTCHK_PROCEED);
+}
+
+/*
+ * Schedule a new request of a given dns object or the generic list
+ */
+
+int
+http_schedule(struct dns_entry *dns, struct uri_list *head)
 {
 	struct uri *uri;
 	struct uri_small *suri, *next;
 	char *host = NULL;
 	u_short port;
-
-	if (connections >= http_maxconnects) {
-		DFPRINTF(2, (stderr, __FUNCTION__": %d >= http_maxconnects\n",
-			     connections));
-		return (-1);
-	}
-
-	if (terminate) {
-		if (connections == 0) {
-			http_savestate();
-			return (-1);
-		}
-
-		fprintf(stdout,
-			"Terminating: %3d of %3d connections pending.\n",
-			connections, http_maxconnects);
-		return (-1);
-	}
+	int res = -1;
 
 	if ((uri = uri_new()) == NULL)
 		return (-1);
 
-	for (suri = TAILQ_FIRST(&uriqueue); suri; suri = next) {
+	for (suri = TAILQ_FIRST(head); suri; suri = next) {
 		next = TAILQ_NEXT(suri, next);
 
-		if (http_hostportfile(suri->uri, &host, &port, NULL) == -1) {
-			fprintf(stderr, __FUNCTION__": illegal url: %s\n",
-				suri->uri);
+		http_hostportfile(suri->uri, &host, &port, NULL);
 
-			TAILQ_REMOVE(&uriqueue, suri, next);
-			uri_small_free(suri);
-			continue;
-		}
+		/* Only up to third outstanding dns queries */
+		if (dns == NULL && dns_connections > http_maxconnects/3)
+			return (-1);
 
 		if (use_robots) {
-			struct http_robots *htrobot;
+			int res = http_robot_check(head, suri, host, port);
 
-			if ((htrobot = http_robots_find(host, port)) == NULL) {
-				if (http_robots_get(host, port) != -1)
-					continue;
-			} else {
-				if (htrobot->flags & HTROBOT_PENDING)
-					continue;
-				if (!http_robots_allow(htrobot, suri->uri)) {
-					fprintf(stderr, __FUNCTION__
-						": robots diallowed url: %s\n",
-						suri->uri);
-				
-					TAILQ_REMOVE(&uriqueue, suri, next);
-					STATS_DECFIFO;
-
-					uri_small_free(suri);
-					continue;
-				}
-			}
 			/* Everything okay */
+			switch (res) {
+			case HTTP_RBTCHK_ASKING:
+				if (connections >= http_maxconnects) {
+					res = 0;
+					goto out;
+				}
+				/* FALLTHROUGH */
+			case HTTP_RBTCHK_PENDING:
+			case HTTP_RBTCHK_DENY:
+				continue;
+			case HTTP_RBTCHK_PROCEED:
+				/* Everything is okay */
+				break;
+			}
 		}
 
-		uri->fd = http_newconnection(uri, host, port);
-		if (uri->fd == -1) {
+		TAILQ_REMOVE(head, suri, next);
+		STATS_DECFIFO;
+
+		/* Copy small to large uri */
+		uri->uri = suri->uri;
+		uri->flags = suri->flags;
+		uri->depth = suri->depth;
+
+		/* suri still needs freeing */
+		suri->uri = NULL;
+
+		if (http_newconnection(uri, host, port) == -1) {
+			/* Swap back to small state */
+			suri->uri = uri->uri;
+			uri->uri = NULL;
+
 			/* These errors might be transient */
 			suri->retry++;
 			if (suri->retry >= HTTP_MAXRETRY) {
-				TAILQ_REMOVE(&uriqueue, suri, next);
 				uri_small_free(suri);
+			} else {
+				http_insert_order(head, suri);
+				STATS_INCFIFO;
 			}
 			continue;
 		}
@@ -616,17 +814,6 @@ http_postevent(void)
 	/* Nothing to do, go away */
 	if (suri == NULL)
 		goto out;
-
-	TAILQ_REMOVE(&uriqueue, suri, next);
-	STATS_DECFIFO;
-
-	/* Copy small to large uri */
-	uri->uri = suri->uri;
-	uri->flags = suri->flags;
-	uri->depth = suri->depth;
-
-	/* Free small uri */
-	suri->uri = NULL;
 	uri_small_free(suri);
 
 	return (0);
@@ -634,14 +821,155 @@ http_postevent(void)
  out:
 	if (uri != NULL)
 		uri_free(uri);
-	return (-1);
+	return (res);
 }
 
 int
-http_add(u_short type, char *url, u_short depth, int bfs)
+dns_schedule(int media)
+{
+	extern struct dns_list dnsqueue;
+	struct uri_list *head;
+	struct uri_small *suri;
+	struct dns_entry *dns;
+	struct timeval now, tv_ready, tv_min;
+	int res = -1;
+	int depth = 0;
+
+	gettimeofday(&now, NULL);
+
+	head = media ? &mediaqueue : &uriqueue;
+	suri = TAILQ_FIRST(head);
+	if (suri != NULL)
+		depth = suri->depth;
+
+	timerclear(&tv_min);
+	TAILQ_FOREACH(dns, &dnsqueue, next) {
+		if (head != NULL && dns->depth < depth) {
+			/* Schedule from generic queue */
+			res = http_schedule(NULL, head);
+			if (res != -1) {
+				dns = NULL;
+				break;
+			}
+
+			/*
+			 * We failed to get one of the generic queue,
+			 * so we may continue with dns entries
+			 */
+			head = NULL;
+		}
+
+		/* Either pending or no time yet */
+		if (!dns_ready(dns, &now, &tv_ready, HTTP_MAXPARALLEL)) {
+			if (timerisset(&tv_ready) &&
+			    (!timerisset(&tv_min) ||
+				timercmp(&tv_ready, &tv_min, <))) {
+				tv_min = tv_ready;
+			}
+			continue;
+		}
+
+		res = http_schedule(dns,
+		    media ? &dns->mediaqueue : &dns->uriqueue);
+		if (res != -1)
+			break;
+	}
+
+	/* Potentially remove an element */
+	if (dns != NULL) {
+		if (res != -1) {
+			http_dns_calcdepth(dns);
+			return (0);
+		}
+	} else if (head != NULL)
+		res = http_schedule(NULL, head);
+
+	if (res == -1 && timerisset(&tv_min) && 
+	    !timeout_pending(&ev_timeout, NULL)) {
+		struct timeval tv;
+		timersub(&tv_min, &now, &tv);
+
+		timeout_add(&ev_timeout, &tv);
+	}
+
+	return (res);
+}
+
+int
+http_postevent(void)
+{
+	int res;
+
+	if (connections >= http_maxconnects) {
+		DFPRINTF(4, (stderr, __FUNCTION__": %d >= http_maxconnects\n",
+			     connections));
+		return (-1);
+	}
+
+	if (terminate) {
+		if (connections == 0) {
+			http_savestate();
+			return (-1);
+		}
+
+		fprintf(stdout,
+			"Terminating: %3d(%3d) of %3d connections pending.\n",
+			connections, dns_connections, http_maxconnects);
+		return (-1);
+	}
+
+	/* We are calling schedule, schedule adds a new timeout if necessary */
+	timeout_del(&ev_timeout);
+
+	/* media list first */
+	res = dns_schedule(1);
+	if (res == 0)
+		return (0);
+
+	/* uri list */
+	res = dns_schedule(0);
+	return (res);
+}
+
+void
+http_insert_order(struct uri_list *head, struct uri_small *uri)
+{
+	struct uri_small *tmp;
+
+	TAILQ_FOREACH(tmp, head, next) {
+		if (tmp->depth <= uri->depth)
+			break;
+	}
+
+	if (tmp == NULL)
+		TAILQ_INSERT_HEAD(head, uri, next);
+	else
+		TAILQ_INSERT_BEFORE(tmp, uri, next);
+}
+
+int
+http_add(u_short type, char *url, u_short depth)
 {
 	struct uri_small *uri;
+	struct uri_list *head;
+	struct dns_entry *dns;
+	u_short port;
+	char *host, *file;
 
+	if (http_hostportfile(url, &host, &port, &file) == -1) {
+		fprintf(stderr, __FUNCTION__": illegal url: %s\n", url);
+		return (-1);
+	}
+
+	dns = dns_find(host);
+	/* On a negative cache hit, we just throw the url away */
+	if (dns != NULL && DNS_NEGATIVE(dns)) {
+		DFPRINTF(1, (stderr, __FUNCTION__
+			     ": dropping %s on negative dns cache hit\n",
+			     url));
+		return (0);
+	}
+	
 	uri = uri_small_new();
 
 	if (uri == NULL) {
@@ -660,14 +988,14 @@ http_add(u_short type, char *url, u_short depth, int bfs)
 	uri->flags = type;
 	uri->depth = depth;
 
-	if (bfs) {
-		TAILQ_INSERT_TAIL(&uriqueue, uri, next);
-	} else {
-		TAILQ_INSERT_HEAD(&uriqueue, uri, next);
-	}
-	STATS_INCFIFO;
+	/* If we have an entry for the hostname queue url up there */
+	if (dns != NULL) {
+		head = isMedia(url) ? &dns->mediaqueue : &dns->uriqueue;
+	} else
+		head = isMedia(url) ? &mediaqueue : &uriqueue;
 
-	http_postevent();
+	http_insert_order(head, uri);
+	STATS_INCFIFO;
 
 	return (0);
 }
@@ -830,7 +1158,8 @@ http_parseheader(struct uri *uri)
 
 	case HTTP_MOVETEMP:
 	case HTTP_MOVEPERM:
-		if (location == NULL || http_movecb == NULL)
+		if (location == NULL || http_movecb == NULL ||
+		    (use_robots && http_isrobotstxt(uri)))
 			goto error;
 
 		p = location;
@@ -861,6 +1190,30 @@ http_parseheader(struct uri *uri)
 		http_dispatch(uri, HTTPDIS_FREE);
 		return;
 	}
+
+	/*
+	 * Check if we have a dispatcher for this format.
+	 * If we not, we are not going to continue with this.
+	 */
+	if (uri->format != NULL) {
+		struct dispatch *dispatch;
+		char *format = uri->format;
+
+		if (use_robots && http_isrobotstxt(uri))
+			goto conturi;
+
+		/* Find a dispatch for this format */
+		TAILQ_FOREACH(dispatch, &dispatchqueue, next) {
+			if (strncasecmp(format, dispatch->type,
+				strlen(dispatch->type)) == 0)
+				break;
+		}
+
+		/* We do not need to download this */
+		if (dispatch == NULL)
+			goto error;
+	}
+ conturi:
 	
 	if (uri->length != -1) {
 		size_t len;
@@ -890,7 +1243,10 @@ http_parseheader(struct uri *uri)
 	return;
 
  error:
-	fprintf(stderr, __FUNCTION__": abort %s: %d\n",	uri->uri, code);
+	http_stats.headerbytes += uri->hdlen;
+
+	DFPRINTF(1, (stderr, __FUNCTION__": abort %s: %d\n",
+		     uri->uri, code));
 	uri_free(uri);
 	return;
 }
@@ -983,6 +1339,19 @@ http_readbody(int fd, short event, void *arg)
 
 int
 http_isrobotstxt(struct uri *uri)
+{
+	if (uri->uri != NULL) {
+		char *file;
+
+		return (http_hostportfile(uri->uri, NULL, NULL, &file) != -1 &&
+		    strcmp(file, "/robots.txt") == 0);
+	}
+
+	return (0);
+}
+
+int
+http_robotstxt(struct uri *uri)
 {
 	char *file;
 
@@ -1010,15 +1379,21 @@ http_dispatch(struct uri *uri, int mustfree)
 	if (format == NULL)
 		format = "text/html";
 
-	http_stats.headerbytes += uri->hdlen;
-	http_stats.bodybytes += uri->bdlen;
+	if (uri->length < HTTP_MAXMEM || uri->bdread == 0) {
+		http_stats.headerbytes += uri->hdlen;
+		http_stats.bodybytes += uri->bdlen;
+	} else {
+		if (uri->bdlen == uri->bdmemlen)
+			http_stats.headerbytes += uri->hdlen;
+		http_stats.bodybytes += uri->bdread;
+	}
 
 	if (uri->flags & HTTP_REQUEST_GET)
 		http_stats.gets++;
 	else
 		http_stats.heads++;
 
-	if (use_robots && http_isrobotstxt(uri)) {
+	if (use_robots && http_robotstxt(uri)) {
 		uri_free(uri);
 		return;
 	}
@@ -1086,12 +1461,14 @@ http_normalize_uri(char *uri)
 	else
 		snprintf(normal, sizeof(normal), "http://%s", host);
 
+	/* Unix file names should not be lowered */
 	p = normal;
 	while (*p)
 		*p++ = tolower(*p);
 
-	/* Unix file names should not be lowered */
-	strlcat(normal, file, sizeof (normal));
+	/* Append file and check for overrun */
+	if (strlcat(normal, file, sizeof (normal)) >= sizeof (normal))
+		return (NULL);
 
 	return (normal);
 }
