@@ -32,12 +32,14 @@
 #define _HTTP_H_
 
 #define HTTP_NAME		"libcrawl"
-#define HTTP_VERSION		"0.2"
+#define HTTP_VERSION		"0.3"
 
 #define HTTP_MAXRETRY		3
 #define HTTP_MAXCONNECTS	20
-#define HTTP_CONNECT_TIMEOUT	30
-#define HTTP_READTIMEOUT	60
+#define HTTP_CONNECT_TIMEOUT	50
+#define HTTP_READTIMEOUT	180
+
+#define HTTP_MAXPARALLEL	3	/* number of connections per host */
 
 #define HTTP_PREFIX		"http://"
 #define HTTP_DEFAULTPORT	80
@@ -83,6 +85,8 @@ struct uri {
 	size_t bdread;
 	size_t bdlen;
 
+	struct dns_entry *dns;	/* reference counter for parallel access */
+
 	int fd;
 	struct event ev;
 };
@@ -99,6 +103,8 @@ struct dispatch {
 	int (*cb)(struct uri *);
 };
 
+TAILQ_HEAD(uri_list, uri_small);
+
 struct uri *uri_new(void);
 void uri_free(struct uri *);
 
@@ -106,10 +112,7 @@ int http_isrobotstxt(struct uri *);
 int http_newconnection(struct uri *, char *, u_short);
 
 int http_init(char *);
-int http_add(u_short, char *, u_short, int);
-
-#define http_add_bfs(x,y,z) http_add(x,y,z,1)
-#define http_add_dfs(x,y,z) http_add(x,y,z,0)
+int http_add(u_short, char *, u_short);
 
 int http_setuseragent(char *);
 
