@@ -46,20 +46,19 @@
 
 #include <event.h>
 
+#include "tree.h"
 #include "connection.h"
 #include "http.h"
 #include "robots.h"
 #include "util.h"
+#include "dns.h"
 
 void http_robots_listfree(char **);
 void http_robots_free(struct http_robots *);
 
-#define ROBOTS_MAXLRUSIZE	1024
+#define ROBOTS_MAXLRUSIZE	32768
 
-TAILQ_HEAD(robots_list, http_robots) robotsqueue;
-int robotsentries;
-
-extern char *rawname;
+SPLAY_HEAD(rtree, http_robots) robottree;
 
 struct stats {
 	size_t robotfiles;
@@ -70,9 +69,26 @@ struct stats {
 	size_t iterations;
 } robot_stats;
 
+static int
+compare(struct http_robots *a, struct http_robots *b)
+{
+	robot_stats.iterations++;
+	return (strcmp(a->host, b->host));
+}
+
+SPLAY_PROTOTYPE(rtree, http_robots, splay_next, compare);
+
+SPLAY_GENERATE(rtree, http_robots, splay_next, compare);
+
+TAILQ_HEAD(robots_list, http_robots) robotsqueue;
+int robotsentries;
+
+extern char *rawname;
+
 void
 http_robots_init(void)
 {
+	SPLAY_INIT(&robottree);
 	TAILQ_INIT(&robotsqueue);
 	robotsentries = 0;
 
@@ -111,6 +127,7 @@ http_robots_new(char *host, u_short port)
 	htrobot->flags = HTROBOT_PENDING;
 
 	TAILQ_INSERT_HEAD(&robotsqueue, htrobot, robot_next);
+	SPLAY_INSERT(rtree, &robottree, htrobot);
 
 	if (robotsentries > ROBOTS_MAXLRUSIZE) {
 		struct http_robots *tmp;
@@ -130,6 +147,7 @@ void
 http_robots_free(struct http_robots *htrobot)
 {
 	TAILQ_REMOVE(&robotsqueue, htrobot, robot_next);
+	SPLAY_REMOVE(rtree, &robottree, htrobot);
 
 	if (htrobot->dislist)
 		http_robots_listfree(htrobot->dislist);
@@ -165,36 +183,31 @@ http_robots_get(char *host, u_short port)
 		return (-1);
 	}
 
-	uri->fd = http_newconnection(uri, host, port);
-	if (uri->fd == -1) {
-		http_robots_free(htrobot);
-		uri_free(uri);
-		return (-1);
-	}
+	if (http_newconnection(uri, host, port) == -1)
+		goto fail;
 
 	uri->flags = HTTP_REQUEST_GET;
-	uri->depth = 0;
+	uri->depth = -1;
 
 	return (0);
+
+ fail:
+	http_robots_free(htrobot);
+	uri_free(uri);
+	return (-1);
 }
 
 struct http_robots *
 http_robots_find(char *host, u_short port)
 {
-	struct http_robots *tmp;
+	struct http_robots *tmp, tmp2;
+
+	tmp2.host = host;
 
 	robot_stats.calls++;
+	tmp = SPLAY_FIND(rtree, &robottree, &tmp2);
 
-	TAILQ_FOREACH(tmp, &robotsqueue, robot_next) {
-		robot_stats.iterations++;
-		if (!strcmp(tmp->host, host) && tmp->port == port) {
-			TAILQ_REMOVE(&robotsqueue, tmp, robot_next);
-			TAILQ_INSERT_HEAD(&robotsqueue, tmp, robot_next);
-			return (tmp);
-		}
-	}
-
-	return (NULL);
+	return (tmp);
 }
 
 int
@@ -285,11 +298,8 @@ http_robots_response(struct uri *uri)
 
 		/* Find line end */
 		lend = strpbrk(p, "\r\n");
-		if (lend == NULL) {
-			fprintf(stderr, __FUNCTION__": illegal robots.txt: %s\n",
-				uri->uri);
-			goto done;
-		}
+		if (lend == NULL)
+			lend = end;
 
 		/* Deal with User-Agent field */
 		if (!strncasecmp(p, FLD_USERAGENT, strlen(FLD_USERAGENT))) {
@@ -380,7 +390,6 @@ http_robots_response(struct uri *uri)
 		}
 	}
 
- done:
 	if (tmplist != NULL) { 
 		if (!(flag & FLG_SKIPDISALLOW)) {
 			/* Insert the current list */
